/*
 * Main.java
 *
 * Created on 27 June 2006, 21:53
 *
 * Copyright  2006 Miles Asvachin
 * Feel free to do whatever you want with this code, but please provide credit
 * where appropriate.
 */

package radixreducer;

import java.awt.datatransfer.*;
import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.math.*;
import javax.swing.border.*;
import javax.swing.text.JTextComponent;

/**
 * Converts numbers between bases.
 *
 * @version 1.2 7 October 2006
 * @author Miles Asvachin
 */
public final class Main extends JScrollPane
{
	private static final int MIN_BASE = 2;
	private static final int MAX_BASE = 36;
	
	private final JTextField[] textFields;
	private final JSpinner precisionSpinner;
	
	private static final int defaultPrecision = 16;
	private static final int MIN_PRECISION = 1;
	private static final int MAX_PRECISION = 128;
	
	private static final String[] baseNames =
	{
		"Binary",
		"Trinary",
		"Quaternary",
		"Quinary",
		"Senary",
		"Septenary",
		"Octal",
		"Nonary",
		"Decimal",
		"Undecimal",
		"Duodecimal",
		"Tridecimal",
		"Quadrodecimal",
		"Quindecimal",
		"Hexadecimal",
		"Septendecimal",
		"Octodecimal",
		"Novemdecimal",
		"Vigesimal",
		"Unvigesimal",
		"Duovigesimal",
		"Trivigesimal",
		"Quadrovigesimal",
		"Quinvigesimal",
		"Sexavigesimal",
		"Septemvigesimal",
		"Octovigesimal",
		"Novemvigesimal",
		"Trigesimal",
		"Untrigesimal",
		"Duotrigesimal",
		"Tritregesimal",
		"Quadrotrigesimal",
		"Quintrigesimal",
		"Sexatrigesimal"
	};
	
	/** Creates a new instance of Main */
	public Main()
	{
		super(VERTICAL_SCROLLBAR_ALWAYS, HORIZONTAL_SCROLLBAR_AS_NEEDED);
		setPreferredSize(new Dimension(400, 529));
		getVerticalScrollBar().setUnitIncrement(33);
		getHorizontalScrollBar().setUnitIncrement(10);
		
		JPanel scrollPanel = new JPanel();
		scrollPanel.setLayout(new BoxLayout(scrollPanel, BoxLayout.Y_AXIS));
		setViewportView(scrollPanel);
		
		//Create precision control.
		JPanel panel = new JPanel(new FlowLayout(FlowLayout.LEFT));
		JLabel label = new JLabel("<html><b>Precision</b></html>");
		panel.add(label);
		
		String[] precisionStrings = new String[MAX_PRECISION - MIN_PRECISION + 1];
		
		for(int i = MIN_PRECISION; i <= MAX_PRECISION; i++)
			precisionStrings[i - MIN_PRECISION] = Integer.toString(i);
		
		precisionSpinner = new JSpinner(new SpinnerListModel(precisionStrings));
		precisionSpinner.setPreferredSize(new Dimension(40, 20));
		precisionSpinner.setToolTipText("Number of places to calculate after the radix point");
		precisionSpinner.setValue(precisionStrings[defaultPrecision - MIN_PRECISION]);
		panel.add(precisionSpinner);
		
		panel.add(Box.createRigidArea(new Dimension(20, 0)));
		
		//Auto-calc check box
		label = new JLabel("<html><b>Auto-calc</b></html>");
		panel.add(label);
		
		final JCheckBox autoCheckBox = new JCheckBox();
		autoCheckBox.setToolTipText("Calculate as number is typed");
		panel.add(autoCheckBox);
		
		setColumnHeaderView(panel);
		
		//Popup menu
		final JPopupMenu popup = new JPopupMenu();
		
		final JMenuItem cutMenuItem = new JMenuItem("Cut");
		cutMenuItem.addActionListener(new ActionListener(){public void actionPerformed(ActionEvent e){((JTextComponent)popup.getInvoker()).cut();}});
		popup.add(cutMenuItem);
		
		final JMenuItem copyMenuItem = new JMenuItem("Copy");
		copyMenuItem.addActionListener(new ActionListener(){public void actionPerformed(ActionEvent e){((JTextComponent)popup.getInvoker()).copy();}});
		popup.add(copyMenuItem);
		
		final JMenuItem pasteMenuItem = new JMenuItem("Paste");
		pasteMenuItem.addActionListener(new ActionListener(){public void actionPerformed(ActionEvent e){((JTextComponent)popup.getInvoker()).paste();}});
		popup.add(pasteMenuItem);
		
		//Create pairs of labels and text fields surrounded by a border, one for each base.
		textFields = new JTextField[MAX_BASE - MIN_BASE + 1];
		
		for(int i = MIN_BASE; i <= MAX_BASE; i++)
		{
			panel = new JPanel();
			panel.setLayout(new BoxLayout(panel, BoxLayout.X_AXIS));
			
			label = new JLabel("<html><b>" + Integer.toString(i) + "</b></html>", JLabel.RIGHT);
			label.setMinimumSize(new Dimension(15, 23));
			label.setPreferredSize(new Dimension(15, 23));
			label.setMaximumSize(new Dimension(15, 23));
			panel.add(label);
			
			JTextField field = new JTextField();
			field.setHorizontalAlignment(JTextField.RIGHT);
			field.setDocument(new RadixDocument(i));
			field.setToolTipText(baseNames[i - MIN_BASE]);
			final JTextField parent = field;
			field.addMouseListener(
					new MouseListener()
					{
						public void mouseClicked(MouseEvent e){}
						public void mouseEntered(MouseEvent e){}
						public void mouseExited(MouseEvent e){}
						public void mousePressed(MouseEvent e){}
						
						public void mouseReleased(MouseEvent e)
						{
							if(!e.isPopupTrigger())
								return;
							
							//Disable cut/copy if no text selected.
							boolean cutCopyEnabled = (parent.getSelectedText() != null);
							cutMenuItem.setEnabled(cutCopyEnabled);
							copyMenuItem.setEnabled(cutCopyEnabled);

							//Disable paste if clipboard doesn't contain any text
							boolean hasText = false;
							try{hasText = Toolkit.getDefaultToolkit().getSystemClipboard().isDataFlavorAvailable(new DataFlavor(String.class, "text/plain"));}
							catch(HeadlessException ex){ex.printStackTrace();}
							pasteMenuItem.setEnabled(hasText);

							popup.show(e.getComponent(), e.getX(), e.getY());
						}
					});
			final int base = i;
			field.addKeyListener(
					new KeyListener()
					{
						public void keyPressed(KeyEvent e){}
						public void keyTyped(KeyEvent e){}
						
						public void keyReleased(KeyEvent e)
						{
							if(e.getKeyCode() != KeyEvent.VK_ENTER && !autoCheckBox.isSelected())
								return;
							
							calculate(base);
						}
					});
			textFields[i - MIN_BASE] = field;
			panel.add(field);
			
			Border border = BorderFactory.createEtchedBorder(EtchedBorder.LOWERED);
			
			//Highlight bases 2, 10 and 16.
			if(i == 2)
				border = BorderFactory.createCompoundBorder(border, BorderFactory.createLineBorder(Color.BLUE, 3));
			else if(i == 10)
				border = BorderFactory.createCompoundBorder(border, BorderFactory.createLineBorder(Color.RED, 3));
			else if(i == 16)
				border = BorderFactory.createCompoundBorder(border, BorderFactory.createLineBorder(new Color(0, 128, 0), 3));
			else
				border = BorderFactory.createCompoundBorder(border, BorderFactory.createLineBorder(Color.LIGHT_GRAY, 3));
			
			panel.setBorder(border);
			
			scrollPanel.add(panel);
		}
	}
	
	public void initFocus()
	{
		if(textFields[0] != null)
			textFields[0].requestFocusInWindow();
	}
	
	/**
	 * @param args the command line arguments
	 */
	public static void main(String[] args)
	{
		javax.swing.SwingUtilities.invokeLater(new Runnable(){public void run(){createAndShowGUI();}});
	}
	
	/**Create GUI.*/
	private static void createAndShowGUI()
	{
		try{UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());}
		catch(ClassNotFoundException ex){ex.printStackTrace();}
		catch(UnsupportedLookAndFeelException ex){ex.printStackTrace();}
		catch(InstantiationException ex){ex.printStackTrace();}
		catch(IllegalAccessException ex){ex.printStackTrace();}
		
		JFrame.setDefaultLookAndFeelDecorated(true);
		
		JFrame frame = new JFrame("Radix Reducer");
		frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		
		Main newContentPane = new Main();
		newContentPane.setOpaque(true);
		frame.setContentPane(newContentPane);
		
		frame.pack();
		newContentPane.initFocus();
		frame.setVisible(true);
	}

	private final void calculate(int base)
	{
		if(base < MIN_BASE || base > MAX_BASE)
			return;
		
		String[] parts = textFields[base - MIN_BASE].getText().split("\\.", 2);
		
		if(parts[0].length() < 1)
			parts[0] = "0";
		
		BigInteger whole = new BigInteger(parts[0], base);
		
		if(parts.length > 1) //Fractional numbers
		{
			int precision = Integer.valueOf(precisionSpinner.getValue().toString());
			BigDecimal fraction = radixFractionToDecimalFraction(parts[1], base, precision);
			
			for(int i = MIN_BASE; i <= MAX_BASE; i++)
			{
				//Don't update the input
				if(i == base)
					continue;
				
				String fractionString;
				String wholeString = whole.toString(i);
				
				try{fractionString = decimalFractionToRadixFraction(fraction, i, precision);}
				catch(FractionOverflowException ex)
				{
					//Round up.
					wholeString = whole.add(new BigInteger("1")).toString();
					fractionString = ".0";
				}

				textFields[i - MIN_BASE].setText(wholeString + fractionString);
			}
		}
		else //Integers are easy.
		{
			for(int i = MIN_BASE; i <= MAX_BASE; i++)
			{
				//Don't update the input
				if(i == base)
					continue;
				
				textFields[i - MIN_BASE].setText(whole.toString(i));
			}
		}
	}
	
	/**
	 * Converts a string representation of a radix fraction in a given base to a radix fraction in decimal.
	 *
	 * @param in String representation of a radix fraction
	 * @param base base of in
	 * @param precision number of radix places to calculate
	 * @return BigDecimal representation of a decimal fraction
	 */
	public static final BigDecimal radixFractionToDecimalFraction(String in, int base, int precision)
	{
		BigDecimal fraction = new BigDecimal(0);
		BigDecimal temp;
		MathContext mc = new MathContext(precision * 2);
		
		for(int i = in.length() - 1; i >= 0; i--)
		{
			temp = new BigDecimal(Long.valueOf(Character.toString(in.charAt(i)), base));
			temp = temp.divide(new BigDecimal(Math.pow(base, i + 1)), mc);
			fraction = fraction.add(temp);
		}
		
		return fraction;
	}
	
	/**
	 * Converts a radix fraction in decimal to a string representation of a radix fraction in a given base.
	 *
	 * @param in BigDecimal representation of a radix fraction
	 * @param base base to convert to
	 * @param precision number of radix places to calculate
	 * @return String representation of a radix fraction
	 */
	public static final String decimalFractionToRadixFraction(BigDecimal in, int base, int precision) throws FractionOverflowException
	{
		StringBuilder result = new StringBuilder(".");
		int nextDigit;
		int i = 0;
		boolean round = false;
		
		while(in.signum() > 0 && i++ < precision)
		{
			in = in.multiply(new BigDecimal(base));

			if(i == precision)
			{
				nextDigit = (int)Math.round(in.doubleValue());
				
				if(nextDigit > base - 1)
				{
					nextDigit = base - 1;
					round = true;
				}
			}
			else
			{
				nextDigit = (int)Math.floor(in.doubleValue());
				
				//Fixes cases where in doesn't fit in a double.
				if(nextDigit > base - 1)
					nextDigit = base - 1;
			}
			
			in = in.subtract(new BigDecimal(nextDigit));
			result.append(Character.forDigit(nextDigit, base));
		}
		
		if(result.charAt(result.length() - 1) == '.')
		{
			result.append('0');
		}
		else
		{
			//Strip trailing zeroes.
			while((result.lastIndexOf("0") == result.length() - 1) && result.length() > 2)
				result.deleteCharAt(result.length() - 1);
		}
		
		//Result needs rounding up.
		if(round)
		{
			for(i = result.length() - 1; i > 0; i--)
			{
				if(Character.digit(result.charAt(i), base) == base - 1)
				{
					result.deleteCharAt(i);
				}
				else
				{
					int rounded = Character.digit(result.charAt(i), base) + 1;
					result.setCharAt(i, Character.forDigit(rounded, base));
					break;
				}
			}
			
			//Result rounds off to exactly one and thus won't fit in a radix fraction.
			if(result.length() < 2)
				throw new FractionOverflowException();
		}
		
		return result.toString();
	}
}